@@ -1,11 +1,9 @@
 from abc import ABC, abstractmethod
-from asyncio import Queue, get_event_loop
+from asyncio import Queue, Task, get_event_loop
 from enum import Enum
-from typing import Any, AsyncGenerator, Dict, List, Union
+from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple, Union
 
 from dagit.templates.playground import TEMPLATE
-from dagster import check
-from dagster.seven import json
 from graphene import Schema
 from graphql.error import GraphQLError
 from graphql.error import format_error as format_graphql_error
@@ -22,6 +20,9 @@ from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse
 from starlette.routing import BaseRoute
 from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState
 
+from dagster import check
+from dagster.seven import json
+
 
 class GraphQLWS(str, Enum):
     PROTOCOL = "graphql-ws"
@@ -66,6 +67,9 @@ class GraphQLServer(ABC):
     def make_request_context(self, conn: HTTPConnection):
         raise NotImplementedError()
 
+    def handle_graphql_errors(self, errors):
+        return [format_graphql_error(err) for err in errors]
+
     async def graphql_http_endpoint(self, request: Request):
         """
         fork of starlette GraphQLApp to allow for
@@ -115,20 +119,14 @@ class GraphQLServer(ABC):
                     status_code=status.HTTP_400_BAD_REQUEST,
                 )
 
-        result = await run_in_threadpool(
-            self._graphql_schema.execute,
-            query,
-            variables=variables,
-            operation_name=operation_name,
-            context=self.make_request_context(request),
-            middleware=self._graphql_middleware,
-        )
+        result = await self.execute_graphql_request(request, query, variables, operation_name)
 
-        error_data = [format_graphql_error(err) for err in result.errors] if result.errors else None
         response_data = {"data": result.data}
-        if error_data:
-            response_data["errors"] = error_data
-        status_code = status.HTTP_400_BAD_REQUEST if result.errors else status.HTTP_200_OK
+        status_code = status.HTTP_200_OK
+
+        if result.errors:
+            response_data["errors"] = self.handle_graphql_errors(result.errors)
+            status_code = status.HTTP_400_BAD_REQUEST
 
         return JSONResponse(response_data, status_code=status_code)
 
@@ -138,9 +136,7 @@ class GraphQLServer(ABC):
         Once we are free of conflicting deps, we should be able to use an impl from
         strawberry-graphql or the like.
         """
-        observables = {}
-        tasks = {}
-        event_loop = get_event_loop()
+        tasks: Dict[str, Task] = {}
 
         await websocket.accept(subprotocol=GraphQLWS.PROTOCOL)
 
@@ -159,60 +155,90 @@ class GraphQLServer(ABC):
                 elif message_type == GraphQLWS.CONNECTION_TERMINATE:
                     await websocket.close()
                 elif message_type == GraphQLWS.START:
-                    try:
-                        data = message["payload"]
-                        query = data["query"]
-                        variables = data.get("variables")
-                        operation_name = data.get("operation_name")
+                    data = message["payload"]
 
-                        request_context = self.make_request_context(websocket)
-                        async_result = self._graphql_schema.execute(
-                            query,
-                            variables=variables,
-                            operation_name=operation_name,
-                            context=request_context,
-                            allow_subscriptions=True,
-                        )
-                    except GraphQLError as error:
-                        payload = format_graphql_error(error)
-                        await _send_message(websocket, GraphQLWS.ERROR, payload, operation_id)
+                    task, error_payload = self.execute_graphql_subscription(
+                        websocket=websocket,
+                        operation_id=operation_id,
+                        query=data["query"],
+                        variables=data.get("variables"),
+                        operation_name=data.get("operation_name"),
+                    )
+                    if error_payload:
+                        await _send_message(websocket, GraphQLWS.ERROR, error_payload, operation_id)
                         continue
 
-                    if isinstance(async_result, ExecutionResult):
-                        if not async_result.errors:
-                            check.failed(
-                                f"Only expect non-async result on error, got {async_result}"
-                            )
-                        payload = format_graphql_error(async_result.errors[0])  # type: ignore
-                        await _send_message(websocket, GraphQLWS.ERROR, payload, operation_id)
-                        continue
+                    assert task is not None
 
-                    # in the future we should get back async gen directly, back compat for now
-                    disposable, async_gen = _disposable_and_async_gen_from_obs(
-                        async_result, event_loop
-                    )
+                    tasks[operation_id] = task
 
-                    observables[operation_id] = disposable
-                    tasks[operation_id] = event_loop.create_task(
-                        _handle_async_results(async_gen, operation_id, websocket)
-                    )
                 elif message_type == GraphQLWS.STOP:
-                    if operation_id not in observables:
+                    if operation_id not in tasks:
                         return
 
-                    observables[operation_id].dispose()
-                    del observables[operation_id]
-
                     tasks[operation_id].cancel()
                     del tasks[operation_id]
 
         except WebSocketDisconnect:
             pass
         finally:
-            for operation_id in observables:
-                observables[operation_id].dispose()
+            for operation_id in tasks:
                 tasks[operation_id].cancel()
 
+    async def execute_graphql_request(
+        self,
+        request: Request,
+        query: str,
+        variables: Optional[Dict[str, Any]],
+        operation_name: Optional[str],
+    ) -> ExecutionResult:
+        # use run_in_threadpool since underlying schema is sync
+        return await run_in_threadpool(
+            self._graphql_schema.execute,
+            query,
+            variables=variables,
+            operation_name=operation_name,
+            context=self.make_request_context(request),
+            middleware=self._graphql_middleware,
+        )
+
+    def execute_graphql_subscription(
+        self,
+        websocket: WebSocket,
+        operation_id: str,
+        query: str,
+        variables: Optional[Dict[str, Any]],
+        operation_name: Optional[str],
+    ) -> Tuple[Optional[Task], Optional[Dict[str, Any]]]:
+        request_context = self.make_request_context(websocket)
+        try:
+            async_result = self._graphql_schema.execute(
+                query,
+                variables=variables,
+                operation_name=operation_name,
+                context=request_context,
+                allow_subscriptions=True,
+            )
+        except GraphQLError as error:
+            error_payload = format_graphql_error(error)
+            return None, error_payload
+
+        if isinstance(async_result, ExecutionResult):
+            if not async_result.errors:
+                check.failed(f"Only expect non-async result on error, got {async_result}")
+            handled_errors = self.handle_graphql_errors(async_result.errors)
+            # return only one entry for subscription response
+            return None, handled_errors[0]
+
+        # in the future we should get back async gen directly, back compat for now
+        disposable, async_gen = _disposable_and_async_gen_from_obs(async_result, get_event_loop())
+        task = get_event_loop().create_task(
+            _handle_async_results(async_gen, operation_id, websocket)
+        )
+        task.add_done_callback(lambda _: disposable.dispose())
+
+        return task, None
+
     def create_asgi_app(
         self,
         **kwargs,
